"""Habitat-related models for lara_django_organisms.

This module contains models related to organism habitats, including
HabitatClass, HabitatParameter, and Habitat.

:authors: mark doerr <mark.doerr@uni-greifswald.de>
"""

import uuid

from django.conf import settings
from django.db import models
from django.utils import timezone

from lara_django_base.models import PhysicalUnit

from .base import ExtraData


class HabitatClass(models.Model):
    """Classes of habitats.

    Examples: maritime, terrestrial, shore, etc.
    """

    model_curi = "lara:organisms/HabitatClass"
    habitatclass_id = models.UUIDField(
        primary_key=True,
        default=uuid.uuid4,
        editable=False,
    )
    name = models.TextField(
        unique=True,
        help_text="maritime, terrestrial, shore",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation",
    )
    description = models.TextField(
        blank=True,
        help_text="description of the habitat class",
    )

    def save(self, *args, **kwargs):
        """Generate default IRI if not provided."""
        if self.iri is None or self.iri == "":
            self.iri = (
                f"{settings.LARA_PREFIXES['lara']}organisms/HabitatClass/"
                f"{self.name.replace(' ', '_').lower().strip()}"
            )

        super().save(*args, **kwargs)

    def __str__(self):
        """Return the name of the habitat class."""
        return self.name or ""

    def __repr__(self):
        """Return the name of the habitat class."""
        return self.name or ""

    class Meta:
        """Meta options for HabitatClass model."""

        verbose_name_plural = "HabitatClasses"


class HabitatParameter(models.Model):
    """Class level Habitat Parameters.

    Can be used to model all kinds of general habitat parameters
    that are common for all organisms of a certain class.
    Habitat parameters of an individual organism should be captured in the
    organism store.

    Examples of these parameters could be:
    - moisture, pH, nutrients, geological properties
    - geo_width_min, geo_width_max
    - temperature_average, temperature_min, temperature_max
    - oxygen_concentration_min, oxygen_concentration_max, oxygen_concentration_average
    - humidity_average
    - pH_average
    - salt concentration
    - moisture_air_min, moisture_air_max, moisture_air_average
    - light_during_day_min, light_during_day_max, light_during_day_average
    - height_min, height_max

    :TODO: parts should be moved to ontology
    """

    model_curi = "lara:organisms/HabitatParameter"
    habitatparameter_id = models.UUIDField(
        primary_key=True,
        default=uuid.uuid4,
        editable=False,
    )
    name_display = models.TextField(
        help_text="human readable display name of the habitat parameter, can contain spaces"
    )
    name = models.TextField(
        unique=True,
        help_text="name of the habitat parameter, e.g. temperature, moisture, pH, ...",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation",
    )
    value_range = models.JSONField(
        null=True,
        blank=True,
        help_text="range of values for this parameter, like min, max, average, ...",
    )
    unit = models.ForeignKey(
        PhysicalUnit,
        on_delete=models.CASCADE,
        related_name="%(app_label)s_%(class)s_unit_related",
        related_query_name="%(app_label)s_%(class)s_unit_related_query",
        blank=True,
        null=True,
        help_text="physical unit of the parameter, e.g. K, mol/L, ...",
    )
    data_extra = models.ManyToManyField(
        ExtraData,
        related_name="%(app_label)s_%(class)s_data_extra_related",
        related_query_name="%(app_label)s_%(class)s_data_extra_related_query",
        blank=True,
        help_text="e.g. more information about this particular parameter ...",
    )
    description = models.TextField(
        null=True,
        blank=True,
        help_text="description of the habitat parameter",
    )
    datetime_last_modified = models.DateTimeField(
        blank=True,
        null=True,
        help_text="date and time when habitat parameter data was last modified",
    )

    def save(self, *args, **kwargs):
        """Generate default values for name and IRI."""
        if self.name is None or self.name == "":
            self.name = self.name_display.replace(" ", "_").lower().strip()

        if self.iri is None or self.iri == "":
            self.iri = (
                f"{settings.LARA_PREFIXES['lara']}organisms/HabitatParameter/"
                f"{self.name.replace(' ', '_').lower().strip()}"
            )

        self.datetime_last_modified = timezone.now()

        super().save(*args, **kwargs)

    def __str__(self):
        """Return the name of the habitat parameter."""
        return self.name or ""

    def __repr__(self):
        """Return the name of the habitat parameter."""
        return self.name or ""


class Habitat(models.Model):
    """Information about the living conditions of an organism.

    Examples: temperature, moisture, pH, etc.
    """

    model_curi = "lara:organisms/Habitat"
    habitat_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name_display = models.TextField(
        help_text="human readable display name of the habitat, can contain spaces"
    )
    name = models.TextField(
        blank=True,
        null=True,
        help_text="habitat name with namespaces, like sea.shore, sea.deepsee, land.vulcan",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation",
    )
    habitat_class = models.ForeignKey(
        HabitatClass,
        on_delete=models.CASCADE,
        related_name="%(app_label)s_%(class)s_habitat_class_related",
        related_query_name="%(app_label)s_%(class)s_habitat_class_related_query",
        blank=True,
        null=True,
        help_text="class of the habitat, like maritime, terrestrial, shore, ...",
    )
    parameters = models.ManyToManyField(
        HabitatParameter,
        blank=True,
        related_name="%(app_label)s_%(class)s_parameters_related",
        related_query_name="%(app_label)s_%(class)s_parameters_related_query",
        help_text="e.g. more parameters, like nutrients, ...",
    )
    description = models.TextField(
        null=True,
        blank=True,
        help_text="description of the habitat",
    )
    datetime_last_modified = models.DateTimeField(
        blank=True,
        null=True,
        help_text="date and time when habitat data was last modified",
    )

    def save(self, *args, **kwargs):
        """Generate default IRI if not provided."""
        if self.iri is None or self.iri == "":
            self.iri = (
                f"{settings.LARA_PREFIXES['lara']}organisms/habitat/"
                f"{self.name.replace(' ', '_').lower().strip()}"
            )

        self.datetime_last_modified = timezone.now()

        super().save(*args, **kwargs)

    def __str__(self):
        """Return the name of the habitat."""
        return self.name or ""

    def __repr__(self):
        """Return the name of the habitat."""
        return self.name or ""



"""_____________________________________________________________________

:PROJECT: LARAsuite

*lara_django_organisms forms *

:details: lara_django_organisms forms module with DaisyUI styling.
         -
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: - run "lara-django-dev forms_generator -c lara_django_organisms > forms.py" to update this file
________________________________________________________________________
"""
# django crispy forms s. https://github.com/django-crispy-forms/django-crispy-forms for []
# generated with django-extensions forms_generator -c  [] > forms.py (LARA-version)

from typing import ClassVar

from django import forms
from django.forms.renderers import TemplatesSetting

from .models import (
    Classis,
    Cultivar,
    Domain,
    ExtraData,
    Familia,
    Genus,
    Habitat,
    HabitatClass,
    HabitatParameter,
    Ordo,
    Organism,
    Phylum,
    Regnum,
    Species,
    Subspecies,
    Trait,
    Variant,
)


class CustomFormRenderer(TemplatesSetting):
    """Custom form renderer for LARA forms."""
    
    field_template_name = "lara_django_organisms/forms/field_group.html"


class SearchableSelectMultiple(forms.SelectMultiple):
    """Custom widget for searchable multi-select using Alpine.js."""
    
    template_name = "lara_django_organisms/widgets/searchable_select_multiple.html"
    
    def __init__(self, attrs=None):
        default_attrs = {"class": "searchable-select-multiple"}
        if attrs:
            default_attrs.update(attrs)
        super().__init__(attrs=default_attrs)


## -------------- ExtraData -------------- ##


class ExtraDataCreateForm(forms.ModelForm):
    """Form for creating new ExtraData with all available fields."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = ExtraData
        fields = (
            "data_type",
            "name",
            "name_full",
            "name_display",
            "namespace",
            "version",
            "hash_sha256",
            "data_json",
            "media_type",
            "iri",
            "url",
            "description",
            "file",
        )

        text_fields_required: ClassVar = ("data_type", "name_display")

        text_fields: ClassVar = (
            "name",
            "name_full",
            "version",
            "hash_sha256",
            "media_type",
        )

        textarea_fields: ClassVar = ("data_json", "description")

        url_fields: ClassVar = ("iri", "url")

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full"})
                for field in text_fields
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            "namespace": forms.Select(attrs={"class": "select select-sm select-bordered w-full"}),
            "file": forms.ClearableFileInput(attrs={"class": "file-input file-input-sm file-input-bordered w-full"}),
        }


class ExtraDataUpdateForm(forms.ModelForm):
    """Form for updating existing ExtraData with all available fields."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = ExtraData
        fields = (
            "data_type",
            "name",
            "name_full",
            "name_display",
            "namespace",
            "version",
            "hash_sha256",
            "data_json",
            "media_type",
            "iri",
            "url",
            "description",
            "file",
        )

        text_fields_required: ClassVar = ("data_type", "name_display")

        text_fields: ClassVar = (
            "name",
            "name_full",
            "version",
            "hash_sha256",
            "media_type",
        )

        textarea_fields: ClassVar = ("data_json", "description")

        url_fields: ClassVar = ("iri", "url")

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full"})
                for field in text_fields
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            "namespace": forms.Select(attrs={"class": "select select-sm select-bordered w-full"}),
            "file": forms.ClearableFileInput(attrs={"class": "file-input file-input-sm file-input-bordered w-full"}),
        }


## -------------- HabitatClass -------------- ##


class HabitatClassCreateForm(forms.ModelForm):
    """Form for creating new HabitatClass."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = HabitatClass
        fields = ("name", "iri", "description")

        text_fields_required: ClassVar = ("name",)

        textarea_fields: ClassVar = ("description",)

        url_fields: ClassVar = ("iri",)

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
        }


class HabitatClassUpdateForm(forms.ModelForm):
    """Form for updating existing HabitatClass."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = HabitatClass
        fields = ("name", "iri", "description")

        text_fields_required: ClassVar = ("name",)

        textarea_fields: ClassVar = ("description",)

        url_fields: ClassVar = ("iri",)

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
        }


## -------------- HabitatParameter -------------- ##


class HabitatParameterCreateForm(forms.ModelForm):
    """Form for creating new HabitatParameter."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = HabitatParameter
        fields = ("name_display", "name", "iri", "value_range", "unit", "data_extra", "description")

        text_fields_required: ClassVar = ("name_display", "name")

        textarea_fields: ClassVar = ("value_range", "description")

        url_fields: ClassVar = ("iri",)

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            "unit": forms.Select(attrs={"class": "select select-sm select-bordered w-full"}),
            "data_extra": SearchableSelectMultiple(),
        }


class HabitatParameterUpdateForm(forms.ModelForm):
    """Form for updating existing HabitatParameter."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = HabitatParameter
        fields = ("name_display", "name", "iri", "value_range", "unit", "data_extra", "description")

        text_fields_required: ClassVar = ("name_display", "name")

        textarea_fields: ClassVar = ("value_range", "description")

        url_fields: ClassVar = ("iri",)

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            "unit": forms.Select(attrs={"class": "select select-sm select-bordered w-full"}),
            "data_extra": SearchableSelectMultiple(),
        }


## -------------- Habitat -------------- ##


class HabitatCreateForm(forms.ModelForm):
    """Form for creating new Habitat."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = Habitat
        fields = ("name_display", "name", "iri", "habitat_class", "parameters", "description")

        text_fields_required: ClassVar = ("name_display",)

        text_fields: ClassVar = ("name",)

        textarea_fields: ClassVar = ("description",)

        url_fields: ClassVar = ("iri",)

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full"})
                for field in text_fields
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            "habitat_class": forms.Select(attrs={"class": "select select-sm select-bordered w-full"}),
            "parameters": SearchableSelectMultiple(),
        }


class HabitatUpdateForm(forms.ModelForm):
    """Form for updating existing Habitat."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = Habitat
        fields = ("name_display", "name", "iri", "habitat_class", "parameters", "description")

        text_fields_required: ClassVar = ("name_display",)

        text_fields: ClassVar = ("name",)

        textarea_fields: ClassVar = ("description",)

        url_fields: ClassVar = ("iri",)

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full"})
                for field in text_fields
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            "habitat_class": forms.Select(attrs={"class": "select select-sm select-bordered w-full"}),
            "parameters": SearchableSelectMultiple(),
        }


## -------------- Trait -------------- ##


class TraitCreateForm(forms.ModelForm):
    """Form for creating new Trait."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = Trait
        fields = ("name_display", "name", "value_range", "characteristics", "unit", "iri", "description")

        text_fields_required: ClassVar = ("name_display", "name")

        textarea_fields: ClassVar = ("value_range", "characteristics", "description")

        url_fields: ClassVar = ("iri",)

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            "unit": forms.Select(attrs={"class": "select select-sm select-bordered w-full"}),
        }


class TraitUpdateForm(forms.ModelForm):
    """Form for updating existing Trait."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = Trait
        fields = ("name_display", "name", "value_range", "characteristics", "unit", "iri", "description")

        text_fields_required: ClassVar = ("name_display", "name")

        textarea_fields: ClassVar = ("value_range", "characteristics", "description")

        url_fields: ClassVar = ("iri",)

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            "unit": forms.Select(attrs={"class": "select select-sm select-bordered w-full"}),
        }


## -------------- Domain -------------- ##


class DomainCreateForm(forms.ModelForm):
    """Form for creating new Domain."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = Domain
        fields = ("name_display", "name", "iri", "traits", "characteristics", "description")

        text_fields_required: ClassVar = ("name_display",)

        text_fields: ClassVar = ("name",)

        textarea_fields: ClassVar = ("characteristics", "description")

        url_fields: ClassVar = ("iri",)

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full"})
                for field in text_fields
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            "traits": SearchableSelectMultiple(),
        }


class DomainUpdateForm(forms.ModelForm):
    """Form for updating existing Domain."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = Domain
        fields = ("name_display", "name", "iri", "traits", "characteristics", "description")

        text_fields_required: ClassVar = ("name_display",)

        text_fields: ClassVar = ("name",)

        textarea_fields: ClassVar = ("characteristics", "description")

        url_fields: ClassVar = ("iri",)

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full"})
                for field in text_fields
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            "traits": SearchableSelectMultiple(),
        }


## -------------- Regnum -------------- ##


class RegnumCreateForm(forms.ModelForm):
    """Form for creating new Regnum."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = Regnum
        fields = ("name_display", "name", "iri", "traits", "characteristics", "description")

        text_fields_required: ClassVar = ("name_display",)

        text_fields: ClassVar = ("name",)

        textarea_fields: ClassVar = ("characteristics", "description")

        url_fields: ClassVar = ("iri",)

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full"})
                for field in text_fields
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            "traits": SearchableSelectMultiple(),
        }


class RegnumUpdateForm(forms.ModelForm):
    """Form for updating existing Regnum."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = Regnum
        fields = ("name_display", "name", "iri", "traits", "characteristics", "description")

        text_fields_required: ClassVar = ("name_display",)

        text_fields: ClassVar = ("name",)

        textarea_fields: ClassVar = ("characteristics", "description")

        url_fields: ClassVar = ("iri",)

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full"})
                for field in text_fields
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            "traits": SearchableSelectMultiple(),
        }


## -------------- Phylum -------------- ##


class PhylumCreateForm(forms.ModelForm):
    """Form for creating new Phylum."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = Phylum
        fields = ("name_display", "name", "name_common", "traits", "characteristics", "iri", "description")

        text_fields_required: ClassVar = ("name_display",)

        text_fields: ClassVar = ("name", "name_common")

        textarea_fields: ClassVar = ("characteristics", "description")

        url_fields: ClassVar = ("iri",)

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full"})
                for field in text_fields
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            "traits": SearchableSelectMultiple(),
        }


class PhylumUpdateForm(forms.ModelForm):
    """Form for updating existing Phylum."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = Phylum
        fields = ("name_display", "name", "name_common", "traits", "characteristics", "iri", "description")

        text_fields_required: ClassVar = ("name_display",)

        text_fields: ClassVar = ("name", "name_common")

        textarea_fields: ClassVar = ("characteristics", "description")

        url_fields: ClassVar = ("iri",)

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full"})
                for field in text_fields
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            "traits": SearchableSelectMultiple(),
        }


## -------------- Classis -------------- ##


class ClassisCreateForm(forms.ModelForm):
    """Form for creating new Classis."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = Classis
        fields = ("name_display", "name", "name_common", "traits", "characteristics", "iri", "description")

        text_fields_required: ClassVar = ("name_display",)

        text_fields: ClassVar = ("name", "name_common")

        textarea_fields: ClassVar = ("characteristics", "description")

        url_fields: ClassVar = ("iri",)

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full"})
                for field in text_fields
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            "traits": SearchableSelectMultiple(),
        }


class ClassisUpdateForm(forms.ModelForm):
    """Form for updating existing Classis."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = Classis
        fields = ("name_display", "name", "name_common", "traits", "characteristics", "iri", "description")

        text_fields_required: ClassVar = ("name_display",)

        text_fields: ClassVar = ("name", "name_common")

        textarea_fields: ClassVar = ("characteristics", "description")

        url_fields: ClassVar = ("iri",)

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full"})
                for field in text_fields
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            "traits": SearchableSelectMultiple(),
        }


## -------------- Ordo -------------- ##


class OrdoCreateForm(forms.ModelForm):
    """Form for creating new Ordo."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = Ordo
        fields = ("name_display", "name", "name_common", "traits", "characteristics", "iri", "description")

        text_fields_required: ClassVar = ("name_display",)

        text_fields: ClassVar = ("name", "name_common")

        textarea_fields: ClassVar = ("characteristics", "description")

        url_fields: ClassVar = ("iri",)

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full"})
                for field in text_fields
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            "traits": SearchableSelectMultiple(),
        }


class OrdoUpdateForm(forms.ModelForm):
    """Form for updating existing Ordo."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = Ordo
        fields = ("name_display", "name", "name_common", "traits", "characteristics", "iri", "description")

        text_fields_required: ClassVar = ("name_display",)

        text_fields: ClassVar = ("name", "name_common")

        textarea_fields: ClassVar = ("characteristics", "description")

        url_fields: ClassVar = ("iri",)

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full"})
                for field in text_fields
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            "traits": SearchableSelectMultiple(),
        }


## -------------- Familia -------------- ##


class FamiliaCreateForm(forms.ModelForm):
    """Form for creating new Familia."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = Familia
        fields = ("name_display", "name", "name_common", "traits", "characteristics", "iri", "description")

        text_fields_required: ClassVar = ("name_display",)

        text_fields: ClassVar = ("name", "name_common")

        textarea_fields: ClassVar = ("characteristics", "description")

        url_fields: ClassVar = ("iri",)

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full"})
                for field in text_fields
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            "traits": SearchableSelectMultiple(),
        }


class FamiliaUpdateForm(forms.ModelForm):
    """Form for updating existing Familia."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = Familia
        fields = ("name_display", "name", "name_common", "traits", "characteristics", "iri", "description")

        text_fields_required: ClassVar = ("name_display",)

        text_fields: ClassVar = ("name", "name_common")

        textarea_fields: ClassVar = ("characteristics", "description")

        url_fields: ClassVar = ("iri",)

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full"})
                for field in text_fields
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            "traits": SearchableSelectMultiple(),
        }


## -------------- Genus -------------- ##


class GenusCreateForm(forms.ModelForm):
    """Form for creating new Genus."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = Genus
        fields = ("name_display", "name", "name_common", "traits", "characteristics", "iri", "description")

        text_fields_required: ClassVar = ("name_display",)

        text_fields: ClassVar = ("name", "name_common")

        textarea_fields: ClassVar = ("characteristics", "description")

        url_fields: ClassVar = ("iri",)

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full"})
                for field in text_fields
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            "traits": SearchableSelectMultiple(),
        }


class GenusUpdateForm(forms.ModelForm):
    """Form for updating existing Genus."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = Genus
        fields = ("name_display", "name", "name_common", "traits", "characteristics", "iri", "description")

        text_fields_required: ClassVar = ("name_display",)

        text_fields: ClassVar = ("name", "name_common")

        textarea_fields: ClassVar = ("characteristics", "description")

        url_fields: ClassVar = ("iri",)

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full"})
                for field in text_fields
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            "traits": SearchableSelectMultiple(),
        }


## -------------- Species -------------- ##


class SpeciesCreateForm(forms.ModelForm):
    """Form for creating new Species."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = Species
        fields = ("name_display", "name", "name_common", "traits", "characteristics", "iri", "description")

        text_fields_required: ClassVar = ("name_display",)

        text_fields: ClassVar = ("name", "name_common")

        textarea_fields: ClassVar = ("characteristics", "description")

        url_fields: ClassVar = ("iri",)

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full"})
                for field in text_fields
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            "traits": SearchableSelectMultiple(),
        }


class SpeciesUpdateForm(forms.ModelForm):
    """Form for updating existing Species."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = Species
        fields = ("name_display", "name", "name_common", "traits", "characteristics", "iri", "description")

        text_fields_required: ClassVar = ("name_display",)

        text_fields: ClassVar = ("name", "name_common")

        textarea_fields: ClassVar = ("characteristics", "description")

        url_fields: ClassVar = ("iri",)

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full"})
                for field in text_fields
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            "traits": SearchableSelectMultiple(),
        }


## -------------- Subspecies -------------- ##


class SubspeciesCreateForm(forms.ModelForm):
    """Form for creating new Subspecies."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = Subspecies
        fields = ("name_display", "name", "name_common", "traits", "characteristics", "iri", "description")

        text_fields_required: ClassVar = ("name_display",)

        text_fields: ClassVar = ("name", "name_common")

        textarea_fields: ClassVar = ("characteristics", "description")

        url_fields: ClassVar = ("iri",)

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full"})
                for field in text_fields
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            "traits": SearchableSelectMultiple(),
        }


class SubspeciesUpdateForm(forms.ModelForm):
    """Form for updating existing Subspecies."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = Subspecies
        fields = ("name_display", "name", "name_common", "traits", "characteristics", "iri", "description")

        text_fields_required: ClassVar = ("name_display",)

        text_fields: ClassVar = ("name", "name_common")

        textarea_fields: ClassVar = ("characteristics", "description")

        url_fields: ClassVar = ("iri",)

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full"})
                for field in text_fields
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            "traits": SearchableSelectMultiple(),
        }


## -------------- Variant -------------- ##


class VariantCreateForm(forms.ModelForm):
    """Form for creating new Variant."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = Variant
        fields = ("name_display", "name", "name_common", "traits", "characteristics", "iri", "description")

        text_fields_required: ClassVar = ("name_display",)

        text_fields: ClassVar = ("name", "name_common")

        textarea_fields: ClassVar = ("characteristics", "description")

        url_fields: ClassVar = ("iri",)

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full"})
                for field in text_fields
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            "traits": SearchableSelectMultiple(),
        }


class VariantUpdateForm(forms.ModelForm):
    """Form for updating existing Variant."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = Variant
        fields = ("name_display", "name", "name_common", "traits", "characteristics", "iri", "description")

        text_fields_required: ClassVar = ("name_display",)

        text_fields: ClassVar = ("name", "name_common")

        textarea_fields: ClassVar = ("characteristics", "description")

        url_fields: ClassVar = ("iri",)

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full"})
                for field in text_fields
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            "traits": SearchableSelectMultiple(),
        }


## -------------- Cultivar -------------- ##


class CultivarCreateForm(forms.ModelForm):
    """Form for creating new Cultivar."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = Cultivar
        fields = ("name_display", "name", "name_common", "traits", "characteristics", "iri", "description")

        text_fields_required: ClassVar = ("name_display",)

        text_fields: ClassVar = ("name", "name_common")

        textarea_fields: ClassVar = ("characteristics", "description")

        url_fields: ClassVar = ("iri",)

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full"})
                for field in text_fields
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            "traits": SearchableSelectMultiple(),
        }


class CultivarUpdateForm(forms.ModelForm):
    """Form for updating existing Cultivar."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = Cultivar
        fields = ("name_display", "name", "name_common", "traits", "characteristics", "iri", "description")

        text_fields_required: ClassVar = ("name_display",)

        text_fields: ClassVar = ("name", "name_common")

        textarea_fields: ClassVar = ("characteristics", "description")

        url_fields: ClassVar = ("iri",)

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full"})
                for field in text_fields
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            "traits": SearchableSelectMultiple(),
        }


## -------------- Organism -------------- ##


class OrganismCreateForm(forms.ModelForm):
    """Form for creating new Organism."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = Organism
        fields = (
            "name_display",
            "name",
            "name_common",
            "iri",
            "pac_id",
            "domain",
            "regnum",
            "phylum",
            "classis",
            "ordo",
            "familia",
            "genus",
            "species",
            "subspecies",
            "variant",
            "cultivar",
            "habitats",
            "organisms",
            "traits",
            "characteristics",
            "genome",
            "hybrid",
            "chimera",
            "description",
            "data_extra",
            "tags",
        )

        text_fields_required: ClassVar = ("name_display",)

        text_fields: ClassVar = ("name", "name_common")

        textarea_fields: ClassVar = ("characteristics", "description")

        url_fields: ClassVar = ("iri", "pac_id", "genome")

        boolean_fields: ClassVar = ("hybrid", "chimera")

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full"})
                for field in text_fields
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            **{
                field: forms.CheckboxInput(attrs={"class": "checkbox checkbox-sm"})
                for field in boolean_fields
            },
            "domain": forms.Select(attrs={"class": "select select-sm select-bordered w-full"}),
            "regnum": forms.Select(attrs={"class": "select select-sm select-bordered w-full"}),
            "phylum": forms.Select(attrs={"class": "select select-sm select-bordered w-full"}),
            "classis": forms.Select(attrs={"class": "select select-sm select-bordered w-full"}),
            "ordo": forms.Select(attrs={"class": "select select-sm select-bordered w-full"}),
            "familia": forms.Select(attrs={"class": "select select-sm select-bordered w-full"}),
            "genus": forms.Select(attrs={"class": "select select-sm select-bordered w-full"}),
            "species": forms.Select(attrs={"class": "select select-sm select-bordered w-full"}),
            "subspecies": forms.Select(attrs={"class": "select select-sm select-bordered w-full"}),
            "variant": forms.Select(attrs={"class": "select select-sm select-bordered w-full"}),
            "cultivar": forms.Select(attrs={"class": "select select-sm select-bordered w-full"}),
            "habitats": SearchableSelectMultiple(),
            "organisms": SearchableSelectMultiple(),
            "traits": SearchableSelectMultiple(),
            "data_extra": SearchableSelectMultiple(),
            "tags": SearchableSelectMultiple(),
        }


class OrganismUpdateForm(forms.ModelForm):
    """Form for updating existing Organism."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:  # noqa: D106
        model = Organism
        fields = (
            "name_display",
            "name",
            "name_common",
            "iri",
            "pac_id",
            "domain",
            "regnum",
            "phylum",
            "classis",
            "ordo",
            "familia",
            "genus",
            "species",
            "subspecies",
            "variant",
            "cultivar",
            "habitats",
            "organisms",
            "traits",
            "characteristics",
            "genome",
            "hybrid",
            "chimera",
            "description",
            "data_extra",
            "tags",
        )

        text_fields_required: ClassVar = ("name_display",)

        text_fields: ClassVar = ("name", "name_common")

        textarea_fields: ClassVar = ("characteristics", "description")

        url_fields: ClassVar = ("iri", "pac_id", "genome")

        boolean_fields: ClassVar = ("hybrid", "chimera")

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full"})
                for field in text_fields
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            **{
                field: forms.CheckboxInput(attrs={"class": "checkbox checkbox-sm"})
                for field in boolean_fields
            },
            "domain": forms.Select(attrs={"class": "select select-sm select-bordered w-full"}),
            "regnum": forms.Select(attrs={"class": "select select-sm select-bordered w-full"}),
            "phylum": forms.Select(attrs={"class": "select select-sm select-bordered w-full"}),
            "classis": forms.Select(attrs={"class": "select select-sm select-bordered w-full"}),
            "ordo": forms.Select(attrs={"class": "select select-sm select-bordered w-full"}),
            "familia": forms.Select(attrs={"class": "select select-sm select-bordered w-full"}),
            "genus": forms.Select(attrs={"class": "select select-sm select-bordered w-full"}),
            "species": forms.Select(attrs={"class": "select select-sm select-bordered w-full"}),
            "subspecies": forms.Select(attrs={"class": "select select-sm select-bordered w-full"}),
            "variant": forms.Select(attrs={"class": "select select-sm select-bordered w-full"}),
            "cultivar": forms.Select(attrs={"class": "select select-sm select-bordered w-full"}),
            "habitats": SearchableSelectMultiple(),
            "organisms": SearchableSelectMultiple(),
            "traits": SearchableSelectMultiple(),
            "data_extra": SearchableSelectMultiple(),
            "tags": SearchableSelectMultiple(),
        }



"""Models package for lara_django_organisms.

This package contains all models organized into logical modules:
- base: Generic/base models (ExtraData, Trait)
- habitat: Habitat-related models (HabitatClass, HabitatParameter, Habitat)
- taxonomy: Taxonomic hierarchy models (Domain, Regnum, Phylum, etc.)
- organisms: Main organism models (Organism, OrganismsSubset, OrderedOrganismsSubset)

:authors: mark doerr <mark.doerr@uni-greifswald.de>
"""

from django.conf import settings

# Import base models
from .base import ExtraData, Trait

# Import habitat models
from .habitat import Habitat, HabitatClass, HabitatParameter

# Import organisms models
from .organisms import OrderedOrganismsSubset, Organism, OrganismsSubset

# Import taxonomy models
from .taxonomy import (
    Classis,
    Cultivar,
    Domain,
    Familia,
    Genus,
    Ordo,
    Phylum,
    Regnum,
    Species,
    Subspecies,
    Variant,
)

# Configure fixtures
settings.FIXTURES += [
    "1_habitat_class_fix",
    "2_domain_fix",
    "3_regnum_fix",
    # "4_phylum_fix",
    # "5_classis_fix",
    # "6_ordo_fix",
    "7_familia_fix",
    "8_genus_fix",
    # "9_species_fix",
    # "10_subspecies_fix",
    # "11_variant_fix",
    # "12_cultivar_fix",
]

__all__ = [
    # Base models
    "ExtraData",
    "Trait",
    # Habitat models
    "HabitatClass",
    "HabitatParameter",
    "Habitat",
    # Taxonomy models
    "Domain",
    "Regnum",
    "Phylum",
    "Classis",
    "Ordo",
    "Familia",
    "Genus",
    "Species",
    "Subspecies",
    "Variant",
    "Cultivar",
    # Organism models
    "Organism",
    "OrganismsSubset",
    "OrderedOrganismsSubset",
]



"""
_____________________________________________________________________.

:PROJECT: LARAsuite

*lara_django_organisms Model Context Protocol (MCP) handler *

:details: lara_django_organisms Model Context Protocol (MCP) handler.

:authors: mark doerr <mark.doerr@uni-greifswald.de>

________________________________________________________________________
"""

from mcp_server import ModelQueryToolset

from .models import (
    Classis,
    Cultivar,
    Domain,
    ExtraData,
    Familia,
    Genus,
    Habitat,
    HabitatClass,
    HabitatParameter,
    OrderedOrganismsSubset,
    Ordo,
    Organism,
    OrganismsSubset,
    Phylum,
    Regnum,
    Species,
    Subspecies,
    Trait,
    Variant,
)


class ExtraDataQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `ExtraData`."""

    model = ExtraData


class HabitatClassQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `HabitatClass`."""

    model = HabitatClass


class HabitatParameterQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `HabitatParameter`."""

    model = HabitatParameter


class HabitatQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `Habitat`."""

    model = Habitat


class TraitQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `Trait`."""

    model = Trait


class DomainQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `Domain`."""

    model = Domain


class RegnumQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `Regnum`."""

    model = Regnum


class PhylumQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `Phylum`."""

    model = Phylum


class ClassisQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `Classis`."""

    model = Classis


class OrdoQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `Ordo`."""

    model = Ordo


class FamiliaQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `Familia`."""

    model = Familia


class GenusQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `Genus`."""

    model = Genus


class SpeciesQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `Species`."""

    model = Species


class SubspeciesQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `Subspecies`."""

    model = Subspecies


class VariantQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `Variant`."""

    model = Variant


class CultivarQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `Cultivar`."""

    model = Cultivar


class OrganismQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `Organism`."""

    model = Organism


class OrganismsSubsetQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `OrganismsSubset`."""

    model = OrganismsSubset


class OrderedOrganismsSubsetQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `OrderedOrganismsSubset`."""

    model = OrderedOrganismsSubset



"""Taxonomy-related models for lara_django_organisms.

This module contains models for the taxonomic hierarchy including
Domain, Regnum, Phylum, Classis, Ordo, Familia, Genus, Species,
Subspecies, Variant, and Cultivar.

:authors: mark doerr <mark.doerr@uni-greifswald.de>
"""

import uuid

from django.conf import settings
from django.db import models
from django.utils import timezone

from .base import Trait


class Domain(models.Model):
    """Taxonomic domain / super kingdom / empire.

    Examples: Archaea, Bacteria, and Eukarya
    """

    model_curi = "lara:organisms/Domain"
    domain_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name_display = models.TextField(
        help_text="human readable display name of the domain, can contain spaces"
    )
    name = models.TextField(
        blank=True,
        null=True,
        help_text="domain name: Archaea, Bacteria, and Eukarya",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation",
    )
    traits = models.ManyToManyField(
        Trait,
        related_name="%(app_label)s_%(class)s_domain_traits_related",
        related_query_name="%(app_label)s_%(class)s_domain_traits",
        blank=True,
        help_text="General domain's traits / features / attributes",
    )
    characteristics = models.JSONField(
        blank=True,
        null=True,
        help_text="General characteristics of Domain. Use 'traits' for specific traits.",
    )
    description = models.TextField(
        null=True,
        blank=True,
        help_text="description of domain",
    )
    datetime_last_modified = models.DateTimeField(
        blank=True,
        null=True,
        help_text="date and time when habitat parameter data was last modified",
    )

    def save(self, *args, **kwargs):
        """Generate default values for name and IRI."""
        if self.name is None or self.name == "":
            self.name = self.name_display.replace(" ", "_").lower().strip()

        if self.iri is None or self.iri == "":
            self.iri = (
                f"{settings.LARA_PREFIXES['lara']}organisms/domain/"
                f"{self.name.replace(' ', '_').lower().strip()}"
            )
        self.datetime_last_modified = timezone.now()
        super().save(*args, **kwargs)

    def __str__(self):
        """Return the name of the domain."""
        return self.name or ""

    def __repr__(self):
        """Return the name of the domain."""
        return self.name or ""


class Regnum(models.Model):
    """Taxonomic Regnum (Kingdom).

    Examples: Animalia, Plantae, Fungi, Protista,
    Archaea/Archaebacteria, and Bacteria/Eubacteria

    Note: might be deprecated
    """

    model_curi = "lara:organisms/Regnum"
    regnum_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name_display = models.TextField(
        help_text="human readable display name of the regnum, can contain spaces"
    )
    name = models.TextField(
        blank=True,
        null=True,
        help_text="Regnum name, like Plantae, Fungi",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation",
    )
    traits = models.ManyToManyField(
        Trait,
        related_name="%(app_label)s_%(class)s_regnum_traits_related",
        related_query_name="%(app_label)s_%(class)s_regnum_traits",
        blank=True,
        help_text="General regnum's traits / features / attributes",
    )
    characteristics = models.JSONField(
        blank=True,
        null=True,
        help_text="General characteristics of Regnum, use 'traits' for specific traits.",
    )
    description = models.TextField(
        null=True,
        blank=True,
        help_text="description of Regnum",
    )
    datetime_last_modified = models.DateTimeField(
        blank=True,
        null=True,
        help_text="date and time when habitat parameter data was last modified",
    )

    def save(self, *args, **kwargs):
        """Generate default values for name and IRI."""
        if self.name is None or self.name == "":
            self.name = self.name_display.replace(" ", "_").lower().strip()

        if self.iri is None or self.iri == "":
            self.iri = (
                f"{settings.LARA_PREFIXES['lara']}organisms/regnum/"
                f"{self.name.replace(' ', '_').lower().strip()}"
            )
        self.datetime_last_modified = timezone.now()
        super().save(*args, **kwargs)

    def __str__(self):
        """Return the name of the regnum."""
        return self.name or ""

    def __repr__(self):
        """Return the name of the regnum."""
        return self.name or ""

    class Meta:
        """Meta options for Regnum model."""

        verbose_name_plural = "regna"


class Phylum(models.Model):
    """Taxonomic Phylum / Devisio / Division."""

    model_curi = "lara:organisms/Phylum"
    phylum_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name_display = models.TextField(
        help_text="human readable display name of the phylum, can contain spaces"
    )
    name = models.TextField(
        blank=True,
        null=True,
        help_text="phylum name, like ",
    )
    name_common = models.TextField(
        blank=True,
        null=True,
        help_text="common phylum name, like ",
    )
    traits = models.ManyToManyField(
        Trait,
        related_name="%(app_label)s_%(class)s_phylum_traits_related",
        related_query_name="%(app_label)s_%(class)s_phylum_traits",
        blank=True,
        help_text="General phylum's traits / features / attributes",
    )
    characteristics = models.JSONField(
        blank=True,
        null=True,
        help_text="Generic characteristics of phylum. Use 'traits' for specific traits.",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation",
    )
    description = models.TextField(
        null=True,
        blank=True,
        help_text="description of phylum",
    )
    datetime_last_modified = models.DateTimeField(
        blank=True,
        null=True,
        help_text="date and time when habitat parameter data was last modified",
    )

    def save(self, *args, **kwargs):
        """Generate default values for name and IRI."""
        if self.name is None or self.name == "":
            self.name = self.name_display.replace(" ", "_").lower().strip()
        if self.iri is None or self.iri == "":
            self.iri = (
                f"{settings.LARA_PREFIXES['lara']}organisms/phylum/"
                f"{self.name.replace(' ', '_').lower().strip()}"
            )
        self.datetime_last_modified = timezone.now()
        super().save(*args, **kwargs)

    def __str__(self):
        """Return the name of the phylum."""
        return self.name or ""

    def __repr__(self):
        """Return the name of the phylum."""
        return self.name or ""

    class Meta:
        """Meta options for Phylum model."""

        verbose_name_plural = "Phyla"


class Classis(models.Model):
    """Taxonomic Classis / Class."""

    model_curi = "lara:organisms/Classis"
    classis_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name_display = models.TextField(
        help_text="human readable display name of the classis, can contain spaces"
    )
    name = models.TextField(
        blank=True,
        null=True,
        help_text="classis name, like ",
    )
    name_common = models.TextField(
        blank=True,
        null=True,
        help_text="common classis name, like ",
    )
    traits = models.ManyToManyField(
        Trait,
        related_name="%(app_label)s_%(class)s_classis_traits_related",
        related_query_name="%(app_label)s_%(class)s_classis_traits",
        blank=True,
        help_text="General classis's traits / features / attributes",
    )
    characteristics = models.JSONField(
        blank=True,
        null=True,
        help_text="generic characteristics of this classis, use 'traits' for specific traits.",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation",
    )
    description = models.TextField(
        null=True,
        blank=True,
        help_text="description of this classis",
    )
    datetime_last_modified = models.DateTimeField(
        blank=True,
        null=True,
        help_text="date and time when habitat parameter data was last modified",
    )

    def save(self, *args, **kwargs):
        """Generate default values for name and IRI."""
        if self.name is None or self.name == "":
            self.name = self.name_display.replace(" ", "_").lower().strip()
        if self.iri is None or self.iri == "":
            self.iri = (
                f"{settings.LARA_PREFIXES['lara']}organisms/classis/"
                f"{self.name.replace(' ', '_').lower().strip()}"
            )
        self.datetime_last_modified = timezone.now()
        super().save(*args, **kwargs)

    def __str__(self):
        """Return the name of the classis."""
        return self.name or ""

    def __repr__(self):
        """Return the name of the classis."""
        return self.name or ""

    class Meta:
        """Meta options for Classis model."""

        verbose_name_plural = "Classes"


class Ordo(models.Model):
    """Taxonomic Ordo / Order."""

    model_curi = "lara:organisms/Ordo"
    ordo_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name_display = models.TextField(
        help_text="human readable display name of the ordo, can contain spaces"
    )
    name = models.TextField(
        blank=True,
        null=True,
        help_text="ordo name, like ",
    )
    name_common = models.TextField(
        blank=True,
        null=True,
        help_text="common ordo name, like ",
    )
    traits = models.ManyToManyField(
        Trait,
        related_name="%(app_label)s_%(class)s_ordo_traits_related",
        related_query_name="%(app_label)s_%(class)s_ordo_traits",
        blank=True,
        help_text="General ordo's traits / features / attributes",
    )
    characteristics = models.JSONField(
        blank=True,
        null=True,
        help_text="Generic characteristics of this ordo, use 'traits' for specific traits.",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation",
    )
    description = models.TextField(
        null=True,
        blank=True,
        help_text="description of this ordo",
    )
    datetime_last_modified = models.DateTimeField(
        blank=True,
        null=True,
        help_text="date and time when habitat parameter data was last modified",
    )

    def save(self, *args, **kwargs):
        """Generate default values for name and IRI."""
        if self.name is None or self.name == "":
            self.name = self.name_display.replace(" ", "_").lower().strip()
        if self.iri is None or self.iri == "":
            self.iri = (
                f"{settings.LARA_PREFIXES['lara']}organisms/ordo/"
                f"{self.name.replace(' ', '_').lower().strip()}"
            )
        self.datetime_last_modified = timezone.now()
        super().save(*args, **kwargs)

    def __str__(self):
        """Return the name of the ordo."""
        return self.name or ""

    def __repr__(self):
        """Return the name of the ordo."""
        return self.name or ""

    class Meta:
        """Meta options for Ordo model."""

        verbose_name_plural = "Ordae"


class Familia(models.Model):
    """Taxonomic Familia / Family."""

    model_curi = "lara:organisms/Familia"
    familia_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name_display = models.TextField(
        help_text="human readable display name of the familia, can contain spaces"
    )
    name = models.TextField(
        blank=True,
        null=True,
        help_text="familia name, like ",
    )
    name_common = models.TextField(
        blank=True,
        null=True,
        help_text="common familia name, like ",
    )
    traits = models.ManyToManyField(
        Trait,
        related_name="%(app_label)s_%(class)s_familia_traits_related",
        related_query_name="%(app_label)s_%(class)s_familia_traits",
        blank=True,
        help_text="General familia's traits / features / attributes",
    )
    characteristics = models.JSONField(
        blank=True,
        null=True,
        help_text="Generic characteristics of this familia, use 'traits' for specific traits.",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation",
    )
    description = models.TextField(
        null=True,
        blank=True,
        help_text="description of this familia",
    )
    datetime_last_modified = models.DateTimeField(
        blank=True,
        null=True,
        help_text="date and time when habitat parameter data was last modified",
    )

    def save(self, *args, **kwargs):
        """Generate default values for name and IRI."""
        if self.name is None or self.name == "":
            self.name = self.name_display.replace(" ", "_").lower().strip()
        if self.iri is None or self.iri == "":
            self.iri = (
                f"{settings.LARA_PREFIXES['lara']}organisms/familia/"
                f"{self.name.replace(' ', '_').lower().strip()}"
            )
        self.datetime_last_modified = timezone.now()
        super().save(*args, **kwargs)

    def __str__(self):
        """Return the name of the familia."""
        return self.name or ""

    def __repr__(self):
        """Return the name of the familia."""
        return self.name or ""

    class Meta:
        """Meta options for Familia model."""

        verbose_name_plural = "Familiae"


class Genus(models.Model):
    """Taxonomic Genus."""

    model_curi = "lara:organisms/Genus"
    genus_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name_display = models.TextField(
        help_text="human readable display name of the genus, can contain spaces"
    )
    name = models.TextField(
        blank=True,
        null=True,
        help_text="genus name, like ",
    )
    name_common = models.TextField(
        blank=True,
        null=True,
        help_text="common genus name, like ",
    )
    traits = models.ManyToManyField(
        Trait,
        related_name="%(app_label)s_%(class)s_genus_traits_related",
        related_query_name="%(app_label)s_%(class)s_genus_traits",
        blank=True,
        help_text="General genus's traits / features / attributes",
    )
    characteristics = models.JSONField(
        blank=True,
        null=True,
        help_text="General characteristics of this genus, use 'traits' for specific traits.",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation",
    )
    description = models.TextField(
        null=True,
        blank=True,
        help_text="description of this genus",
    )
    datetime_last_modified = models.DateTimeField(
        blank=True,
        null=True,
        help_text="date and time when habitat parameter data was last modified",
    )

    def save(self, *args, **kwargs):
        """Generate default values for name and IRI."""
        if self.name is None or self.name == "":
            self.name = self.name_display.replace(" ", "_").lower().strip()
        if self.iri is None or self.iri == "":
            self.iri = (
                f"{settings.LARA_PREFIXES['lara']}organisms/genus/"
                f"{self.name.replace(' ', '_').lower().strip()}"
            )
        self.datetime_last_modified = timezone.now()
        super().save(*args, **kwargs)

    def __str__(self):
        """Return the name of the genus."""
        return self.name or ""

    def __repr__(self):
        """Return the name of the genus."""
        return self.name or ""

    class Meta:
        """Meta options for Genus model."""

        verbose_name_plural = "Genera"


class Species(models.Model):
    """Taxonomic Species."""

    model_curi = "lara:organisms/Species"
    species_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name_display = models.TextField(
        help_text="human readable display name of the species, can contain spaces"
    )
    name = models.TextField(
        blank=True,
        null=True,
        help_text="species name, like ",
    )
    name_common = models.TextField(
        blank=True,
        null=True,
        help_text="common species name, like ",
    )
    traits = models.ManyToManyField(
        Trait,
        related_name="%(app_label)s_%(class)s_species_traits_related",
        related_query_name="%(app_label)s_%(class)s_species_traits",
        blank=True,
        help_text="General species's traits / features / attributes",
    )
    characteristics = models.JSONField(
        blank=True,
        null=True,
        help_text="General characteristics of this species, use 'traits' for specific traits.",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation",
    )
    description = models.TextField(
        null=True,
        blank=True,
        help_text="description of this species",
    )
    datetime_last_modified = models.DateTimeField(
        blank=True,
        null=True,
        help_text="date and time when habitat parameter data was last modified",
    )

    def save(self, *args, **kwargs):
        """Generate default values for name and IRI."""
        if self.name is None or self.name == "":
            self.name = self.name_display.replace(" ", "_").lower().strip()
        if self.iri is None or self.iri == "":
            self.iri = (
                f"{settings.LARA_PREFIXES['lara']}organisms/species/"
                f"{self.name.replace(' ', '_').lower().strip()}"
            )
        self.datetime_last_modified = timezone.now()
        super().save(*args, **kwargs)

    def __str__(self):
        """Return the name of the species."""
        return self.name or ""

    def __repr__(self):
        """Return the name of the species."""
        return self.name or ""

    class Meta:
        """Meta options for Species model."""

        verbose_name_plural = "Speciae"


class Subspecies(models.Model):
    """Taxonomic Subspecies."""

    model_curi = "lara:organisms/Subspecies"
    subspecies_id = models.UUIDField(
        primary_key=True,
        default=uuid.uuid4,
        editable=False,
    )
    name_display = models.TextField(
        help_text="human readable display name of the subspecies, can contain spaces"
    )
    name = models.TextField(
        blank=True,
        null=True,
        help_text="subspecies name, like ",
    )
    name_common = models.TextField(
        blank=True,
        null=True,
        help_text="common subspecies name, like ",
    )
    traits = models.ManyToManyField(
        Trait,
        related_name="%(app_label)s_%(class)s_subspecies_traits_related",
        related_query_name="%(app_label)s_%(class)s_subspecies_traits",
        blank=True,
        help_text="General subspecies's traits / features / attributes",
    )
    characteristics = models.JSONField(
        blank=True,
        null=True,
        help_text="General characteristics of this subspecies, use 'traits' for specific traits.",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation",
    )
    description = models.TextField(
        null=True,
        blank=True,
        help_text="description of this subspecies",
    )
    datetime_last_modified = models.DateTimeField(
        blank=True,
        null=True,
        help_text="date and time when habitat parameter data was last modified",
    )

    def save(self, *args, **kwargs):
        """Generate default values for name and IRI."""
        if self.name is None or self.name == "":
            self.name = self.name_display.replace(" ", "_").lower().strip()
        if self.iri is None or self.iri == "":
            self.iri = (
                f"{settings.LARA_PREFIXES['lara']}organisms/subspecies/"
                f"{self.name.replace(' ', '_').lower().strip()}"
            )
        self.datetime_last_modified = timezone.now()
        super().save(*args, **kwargs)

    def __str__(self):
        """Return the name of the subspecies."""
        return self.name or ""

    def __repr__(self):
        """Return the name of the subspecies."""
        return self.name or ""

    class Meta:
        """Meta options for Subspecies model."""

        verbose_name_plural = "Subspecieae"


class Variant(models.Model):
    """Taxonomic Variant."""

    model_curi = "lara:organisms/Variant"
    variant_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name_display = models.TextField(
        help_text="human readable display name of the variant, can contain spaces"
    )
    name = models.TextField(
        blank=True,
        null=True,
        help_text="variant name, like ",
    )
    name_common = models.TextField(
        blank=True,
        null=True,
        help_text="common variant name, like ",
    )
    traits = models.ManyToManyField(
        Trait,
        related_name="%(app_label)s_%(class)s_variant_traits_related",
        related_query_name="%(app_label)s_%(class)s_variant_traits",
        blank=True,
        help_text="General variant's traits / features / attributes",
    )
    characteristics = models.JSONField(
        blank=True,
        null=True,
        help_text="General characteristics of this variant, use 'traits' for specific traits.",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation",
    )
    description = models.TextField(
        null=True,
        blank=True,
        help_text="description of this variant",
    )
    datetime_last_modified = models.DateTimeField(
        blank=True,
        null=True,
        help_text="date and time when habitat parameter data was last modified",
    )

    def save(self, *args, **kwargs):
        """Generate default values for name and IRI."""
        if self.name is None or self.name == "":
            self.name = self.name_display.replace(" ", "_").lower().strip()
        if self.iri is None or self.iri == "":
            self.iri = (
                f"{settings.LARA_PREFIXES['lara']}organisms/variant/"
                f"{self.name.replace(' ', '_').lower().strip()}"
            )
        self.datetime_last_modified = timezone.now()
        super().save(*args, **kwargs)

    def __str__(self):
        """Return the name of the variant."""
        return self.name or ""

    def __repr__(self):
        """Return the name of the variant."""
        return self.name or ""


class Cultivar(models.Model):
    """Taxonomic Cultivar / Plant variety originating from cultivation / cell line."""

    model_curi = "lara:organisms/Cultivar"
    cultivar_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name_display = models.TextField(
        help_text="human readable display name of the cultivar, can contain spaces"
    )
    name = models.TextField(
        blank=True,
        null=True,
        help_text="cultivar name, like ",
    )
    name_common = models.TextField(
        blank=True,
        null=True,
        help_text="common cultivar name, like ",
    )
    traits = models.ManyToManyField(
        Trait,
        related_name="%(app_label)s_%(class)s_cultivar_traits_related",
        related_query_name="%(app_label)s_%(class)s_cultivar_traits",
        blank=True,
        help_text="General cultivar's traits / features / attributes",
    )
    characteristics = models.JSONField(
        blank=True,
        null=True,
        help_text="General characteristics of this cultivar, use 'traits' for specific traits.",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation",
    )
    description = models.TextField(
        null=True,
        blank=True,
        help_text="description of this cultivar",
    )
    datetime_last_modified = models.DateTimeField(
        blank=True,
        null=True,
        help_text="date and time when habitat parameter data was last modified",
    )

    def save(self, *args, **kwargs):
        """Generate default values for name and IRI."""
        if self.name is None or self.name == "":
            self.name = self.name_display.replace(" ", "_").lower().strip()
        if self.iri is None or self.iri == "":
            self.iri = (
                f"{settings.LARA_PREFIXES['lara']}organisms/cultivar/"
                f"{self.name.replace(' ', '_').lower().strip()}"
            )
        self.datetime_last_modified = timezone.now()
        super().save(*args, **kwargs)

    def __str__(self):
        """Return the name of the cultivar."""
        return self.name or ""

    def __repr__(self):
        """Return the name of the cultivar."""
        return self.name or ""



"""Organism-related models for lara_django_organisms.

This module contains the main Organism model and related subset models.

:authors: mark doerr <mark.doerr@uni-greifswald.de>
"""

import logging
import uuid

from django.conf import settings
from django.db import models
from django.utils import timezone

from lara_django_base.models import Tag
from lara_django_people.models import ItemSubsetAbstr

from .base import ExtraData, Trait
from .habitat import Habitat
from .taxonomy import (
    Classis,
    Cultivar,
    Domain,
    Familia,
    Genus,
    Ordo,
    Phylum,
    Regnum,
    Species,
    Subspecies,
    Variant,
)

logger = logging.getLogger(__name__)


class Organism(models.Model):
    """Organism / species / strain or cell line.

    Details should be in the organism's ontology.
    """

    model_curi = "lara:organisms/Organism"
    organism_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name_display = models.TextField(
        help_text="human readable display name of the organism, can contain spaces"
    )
    name = models.TextField(
        blank=True,
        null=True,
        help_text="organism name for referencing in code, should not contain whitespaces",
    )
    name_common = models.TextField(
        blank=True,
        null=True,
        help_text="common organism name, like 'elephant' ",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation",
    )
    # PAC-ID:  https://github.com/ApiniLabs/PAC-ID
    pac_id = models.URLField(
        blank=True,
        null=True,
        unique=True,
        help_text=(
            "Publicly Addressable Content IDentifier is a globally unique identifier "
            "that operates independently of a central registry."
        ),
    )
    # taxonomy / classification
    domain = models.ForeignKey(
        Domain,
        on_delete=models.CASCADE,
        related_name="%(app_label)s_%(class)s_organism_domain_related",
        related_query_name="%(app_label)s_%(class)s_organism_domain",
        blank=True,
        null=True,
        help_text="organism's domain name",
    )
    regnum = models.ForeignKey(
        Regnum,
        on_delete=models.CASCADE,
        related_name="%(app_label)s_%(class)s_organism_regnum_related",
        related_query_name="%(app_label)s_%(class)s_organism_regnum",
        blank=True,
        null=True,
        help_text="organism's regnum name",
    )
    phylum = models.ForeignKey(
        Phylum,
        on_delete=models.CASCADE,
        related_name="%(app_label)s_%(class)s_organism_phylum_related",
        related_query_name="%(app_label)s_%(class)s_organism_phylum",
        blank=True,
        null=True,
        help_text="organism's phylum name",
    )
    classis = models.ForeignKey(
        Classis,
        on_delete=models.CASCADE,
        related_name="%(app_label)s_%(class)s_organism_classis_related",
        related_query_name="%(app_label)s_%(class)s_organism_classis",
        blank=True,
        null=True,
        help_text="organism's classis (class) name",
    )
    ordo = models.ForeignKey(
        Ordo,
        on_delete=models.CASCADE,
        related_name="%(app_label)s_%(class)s_organism_ordo_related",
        related_query_name="%(app_label)s_%(class)s_organism_ordo",
        blank=True,
        null=True,
        help_text="organism's ordo name",
    )
    familia = models.ForeignKey(
        Familia,
        on_delete=models.CASCADE,
        related_name="%(app_label)s_%(class)s_organism_familia_related",
        related_query_name="%(app_label)s_%(class)s_organism_familia",
        blank=True,
        null=True,
        help_text="organism's familia (family) name",
    )
    genus = models.ForeignKey(
        Genus,
        on_delete=models.CASCADE,
        related_name="%(app_label)s_%(class)s_organism_genus_related",
        related_query_name="%(app_label)s_%(class)s_organism_genus",
        blank=True,
        null=True,
        help_text="organism's genus name",
    )
    species = models.ForeignKey(
        Species,
        on_delete=models.CASCADE,
        related_name="%(app_label)s_%(class)s_organism_species_related",
        related_query_name="%(app_label)s_%(class)s_organism_species",
        blank=True,
        null=True,
        help_text="organism's species name",
    )
    subspecies = models.ForeignKey(
        Subspecies,
        on_delete=models.CASCADE,
        related_name="%(app_label)s_%(class)s_organism_subspecies_related",
        related_query_name="%(app_label)s_%(class)s_organism_subspecies",
        blank=True,
        null=True,
        help_text="organism's subspecies name",
    )
    variant = models.ForeignKey(
        Variant,
        on_delete=models.CASCADE,
        related_name="%(app_label)s_%(class)s_organism_variant_related",
        related_query_name="%(app_label)s_%(class)s_organism_variant",
        blank=True,
        null=True,
        help_text="organism's varian name",
    )
    cultivar = models.ForeignKey(
        Cultivar,
        on_delete=models.CASCADE,
        related_name="%(app_label)s_%(class)s_organism_cultivar_related",
        related_query_name="%(app_label)s_%(class)s_organism_cultivar",
        blank=True,
        null=True,
        help_text="organism's cultivar name",
    )
    habitats = models.ManyToManyField(
        Habitat,
        related_name="%(app_label)s_%(class)s_organism_habitat_related",
        related_query_name="%(app_label)s_%(class)s_organism_habitat",
        blank=True,
        help_text="general organism's habitats",
    )
    organisms = models.ManyToManyField(
        "self",
        related_query_name="%(app_label)s_%(class)s_organism_organisms",
        blank=True,
        help_text="organisms involved in hybrid or chimera",
    )
    traits = models.ManyToManyField(
        Trait,
        related_name="%(app_label)s_%(class)s_organism_traits_related",
        related_query_name="%(app_label)s_%(class)s_organism_traits",
        blank=True,
        help_text="general organism's traits / features / attributes",
    )
    characteristics = models.JSONField(
        blank=True,
        null=True,
        help_text="General characteristics of this variant, use 'traits' for specific traits.",
    )
    # consider a link to lara-django-sequences to store the genome as a sequence
    # efficiency needs to be tested
    genome = models.URLField(
        blank=True,
        null=True,
        max_length=512,
        help_text="URL to organism's (reference) genome",
    )

    hybrid = models.BooleanField(
        default=False,
        help_text=(
            "Hybrid - organism with mixed genetic material "
            "like mule, liger/tigon, plant hybrids."
        ),
    )
    chimera = models.BooleanField(
        default=False,
        help_text=(
            "Chimera - organisms with mixed cells from different species, "
            "like apple/pear tree."
        ),
    )

    description = models.TextField(
        null=True,
        blank=True,
        help_text="description of the organism",
    )

    data_extra = models.ManyToManyField(
        ExtraData,
        related_name="%(app_label)s_%(class)s_organism_extradata_related",
        related_query_name="%(app_label)s_%(class)s_organism_extradata",
        blank=True,
        help_text="organism related extra ",
    )
    tags = models.ManyToManyField(
        Tag,
        related_name="%(app_label)s_%(class)s_organism_tags_related",
        related_query_name="%(app_label)s_%(class)s_organism_tags",
        blank=True,
        help_text="organism related tags",
    )
    datetime_last_modified = models.DateTimeField(
        blank=True,
        null=True,
        help_text="date and time when organism data was last modified",
    )

    def save(self, *args, **kwargs):
        """Generate default values for name_display, name, and IRI."""
        if self.name_display is None or self.name_display == "":
            try:
                self.name_display = f"{self.genus.name.capitalize()} {self.species.name}"
            except Exception as e:
                logger.error(f"Error in generating name_display for organism: {e}")

        if self.name is None or self.name == "":
            if self.name_display is not None and self.name_display != "":
                self.name = self.name_display.replace(" ", "_").lower().strip()
            else:
                try:
                    self.name = f"{self.genus.name}_{self.species.name}".lower().strip()
                except Exception as e:
                    logger.error(f"Error in generating name for organism: {e}")

        if self.iri is None or self.iri == "":
            self.iri = (
                f"{settings.LARA_PREFIXES['lara']}organisms/organism/"
                f"{self.name.replace(' ', '_').lower().strip()}"
            )

        self.datetime_last_modified = timezone.now()

        super().save(*args, **kwargs)

    def __str__(self):
        """Return the name of the organism."""
        return self.name or ""

    def __repr__(self):
        """Return the name of the organism."""
        return self.name or ""

    class Meta:
        """Meta options for Organism model."""

        constraints = [
            models.UniqueConstraint(
                fields=[
                    "name",
                    "genus",
                    "species",
                    "subspecies",
                    "variant",
                    "cultivar",
                ],
                name="unique_organism",
            )
        ]


class OrganismsSubset(ItemSubsetAbstr):
    """User or group defined subset of organisms."""

    model_curi = "lara:organisms/OrganismsSubset"
    organismssubset_id = models.UUIDField(
        primary_key=True,
        default=uuid.uuid4,
        editable=False,
    )
    organisms = models.ManyToManyField(
        Organism,
        related_name="%(app_label)s_%(class)s_organisms_related",
        related_query_name="%(app_label)s_%(class)s_organisms",
        blank=True,
        help_text="organisms in this subset",
        through="OrderedOrganismsSubset",
    )


class OrderedOrganismsSubset(models.Model):
    """Through model for OrganismsSubset and Organism relationship.

    This class can be used to store the order of organisms in a subset.
    """

    orderedorganismssubset_id = models.UUIDField(
        primary_key=True,
        default=uuid.uuid4,
        editable=False,
    )

    organism = models.ForeignKey(
        Organism,
        related_name="%(app_label)s_%(class)s_organism_related",
        related_query_name="%(app_label)s_%(class)s_organism",
        on_delete=models.CASCADE,
        help_text="organism in the subset",
    )

    organisms_subset = models.ForeignKey(
        OrganismsSubset,
        related_name="%(app_label)s_%(class)s_organismssubset_related",
        related_query_name="%(app_label)s_%(class)s_organismssubset",
        on_delete=models.CASCADE,
        help_text="subset of organisms",
    )

    order = models.IntegerField(help_text="order of organism in subset")

    class Meta:
        """Meta options for OrderedOrganismsSubset model."""

        constraints = [
            models.UniqueConstraint(
                fields=["organism", "organisms_subset"],
                name="%(app_label)s_%(class)s_organisms_organisms_subset_unique",
            ),
        ]
        ordering = ("order",)



"""Base models for lara_django_organisms.

This module contains generic/base model classes that are used across
the organisms app, including ExtraData and Trait models.

:authors: mark doerr <mark.doerr@uni-greifswald.de>
"""

import hashlib
import json
import logging
import uuid

from django.conf import settings
from django.db import models
from django.utils import timezone

from lara_django_base.models import ExtraDataAbstr, PhysicalUnit

logger = logging.getLogger(__name__)


class ExtraData(ExtraDataAbstr):
    """Extended data model for additional organism information.

    This class can be used to extend data by extra information,
    e.g. more telephone numbers, customer numbers, files, etc.
    """

    model_curi = "lara:organisms/ExtraData"
    file = models.FileField(
        upload_to="organisms",
        blank=True,
        null=True,
        help_text="rel. path/filename",
    )

    def save(self, *args, **kwargs):
        """Generate default values for version, hash, name, and IRI."""
        if self.version is None or self.version == "":
            self.version = "0.0.1"

        if self.hash_sha256 is None:
            if self.data_json is not None:
                to_hash = json.dumps(self.data_json)
                self.hash_sha256 = hashlib.sha256(to_hash.encode("utf-8")).hexdigest()
            else:
                self.hash_sha256 = hashlib.sha256(
                    str(self.extradata_id).encode("utf-8")
                ).hexdigest()

        if self.name is None or self.name == "":
            if self.name_display is None or self.name_display == "":
                self.name = "_".join(
                    (
                        timezone.now().strftime("%Y%m%d_%H%M%S"),
                        "data",
                        self.hash_sha256[:8],
                    )
                )
            else:
                self.name = self.name_display.replace(" ", "_").lower().strip()
        else:
            self.name = self.name.replace(" ", "_").lower().strip()

        if self.name_full is None or self.name_full == "":
            if self.namespace is not None and self.version is not None:
                self.name_full = (
                    f"{self.namespace.uri}/"
                    + "_".join((self.name.replace(" ", "_"), self.version))
                )
            else:
                self.name_full = "_".join((self.name.replace(" ", "_"), self.version))

        if not self.iri:
            netloc = self.namespace.parse_URI().netloc
            path = self.namespace.parse_URI().path
            self.iri = (
                f"{settings.LARA_PREFIXES['lara']}organisms/ExtraData/"
                f"{netloc}{path}/{self.name.replace(' ', '_').lower().strip()}"
            )

        self.datetime_last_modified = timezone.now()

        super().save(*args, **kwargs)


class Trait(models.Model):
    """Phenotypical, phenomenological or genetic traits of organisms.

    Examples include: branched, round, oval, GRAM staining, genetic fingerprint sequence,
    antibiotic resistance, morphology, size, shape, motility, spore formation, flagella,
    capsule presence, growth optima (temp, pH, salt, sugar), doubling rate, metabolites,
    nutrients, metabolic characteristics, pathogenicity, membrane potential, legal status,
    culture conditions, etc.
    """

    model_curi = "lara:organisms/Trait"
    trait_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name_display = models.TextField(
        help_text="human readable display name of the trait, can contain spaces"
    )
    name = models.TextField(
        unique=True,
        help_text=(
            "Name of the trait / feature attribute of the organisms, "
            "like branched, round, oval, staining, genetic fingerprint sequence"
        ),
    )
    value_range = models.JSONField(
        null=True,
        blank=True,
        help_text="range of values for this trait, like min, max, average, ...",
    )
    characteristics = models.JSONField(
        blank=True,
        null=True,
        help_text="characteristics of this trait",
    )
    unit = models.ForeignKey(
        PhysicalUnit,
        on_delete=models.CASCADE,
        related_name="%(app_label)s_%(class)s_unit_related",
        related_query_name="%(app_label)s_%(class)s_unit_related_query",
        blank=True,
        null=True,
        help_text="physical unit of the parameter, e.g. m, m^2, V, ...",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation",
    )
    description = models.TextField(
        null=True,
        blank=True,
        help_text="description of the feature class",
    )
    datetime_last_modified = models.DateTimeField(
        blank=True,
        null=True,
        help_text="date and time when trait data was last modified",
    )

    def save(self, *args, **kwargs):
        """Generate default values for name and IRI."""
        if self.name is None or self.name == "":
            self.name = self.name_display.replace(" ", "_").lower().strip()

        if self.iri is None or self.iri == "":
            self.iri = (
                f"{settings.LARA_PREFIXES['lara']}organisms/FeatureClass/"
                f"{self.name.replace(' ', '_').lower().strip()}"
            )

        self.datetime_last_modified = timezone.now()

        super().save(*args, **kwargs)

    def __str__(self):
        """Return the name of the trait."""
        return self.name or ""

    def __repr__(self):
        """Return the name of the trait."""
        return self.name or ""

    class Meta:
        """Meta options for Trait model."""

        verbose_name_plural = "FeatureClasses"
